# -*- coding: utf-8 -*-
#========================#
# 頻度計算プログラム     #
# freq_analysis.rb       #
# ver. 2018.03           #
# 第一引数は塩基配列     #
# 第二引数は塩基パターン #
# 塩基配列末尾の処理     #
# : 不足分を末尾で調整   #
#========================#

$:.unshift(File.dirname(__FILE__))

##########################################################

### 以下 設定部分 ###


# (1) 断片化サイズとステップサイズの設定
# WINDOW_SIZE : 塩基配列を断片化する配列長(断片化サイズ)
# STEP_SIZE   : 次の頻度計算を行う開始位置(ステップサイズ)
# READ_SIZE   : 頻度計算を行う配列長 
# 入力された塩基配列の配列長が断片化サイズ以下の場合は
# 頻度計算を行いません
WINDOW_SIZE = 5000
STEP_SIZE = WINDOW_SIZE
READ_SIZE = WINDOW_SIZE

# (2) Nの許容範囲の設定
# 入力された塩基配列中に
# Nが設定された割合以上含まれている場合は
# 頻度計算を行いません
# 100% : 1.0
#  10% : 0.1
# 0.1% : 0.01
NPAR = 0.1

# (3) 出力される出現頻度ファイルの拡張子の設定
end_name ='.freq'

# (4) 計算方法の設定
# 100分率頻度:100hindo
# 1分率頻度  :hindo
# 度数計算   :dosu
CALC = 'dosu'


### 以上 設定部分###

###########################################################




##########(main_method)##########
def oligonucleotide_freq(flat_file_name, pattern_file_name, end_name = "")
  print "#{flat_file_name}の頻度解析プログラムを開始します\n"
  print "#{pattern_file_name}をパターンファイルとして用います\n"

 #初期化 
  flat_name = File.basename(flat_file_name)
  dna_seq = ""
  word_len = 0
  dimension = 0
  word_pattern = []
  first_line = ""
  second_line = ""
  same_pattern = []
  

  #fastaデータの読み込み
  File.open(flat_file_name){|flat_data|
    flat_data.each{|line_data|
      if line_data =~ /^>/ 
        first_line = line_data.chomp
      else
        dna_seq << line_data.chomp
      end
    }
  }
  
  
  #DegeTetra.ptn(パターンファイルの読み込み)
  File.open(pattern_file_name){|pattern|
    word_len = pattern.gets.chomp.to_i   # 塩基の長さ(ここでは4)
    dimension = pattern.gets.chomp.to_i  # 次元数(ここでは136)

    pattern.each{|line|
      word_pattern.push(line.chomp)      #全パターンを配列に格納


      if line =~ /\+/
        column = line.split(/\+/)      
        pattern_left = column[0]
        pattern_right = column[1].chomp
        
        #パターンファイルが同じ場合
        if pattern_left == pattern_right
          same_pattern.push(pattern_left) #同じパターンファイルを配列に格納
        end 
      end
    }
  }


  dna_seq.upcase!   # 全ての文字を大文字にする。小文字込みの断片への対応
   
  dna_seq_len = dna_seq.length       # 頻度解析対象となる塩基配列の長さ
  
  
  #=====これ以下の配列長は処理しない=====
  if dna_seq_len >= READ_SIZE
  #======================================
    
    
    #塩基配列を断片化させる
    dna_seq_frag_ary = fragmentation(dna_seq, dna_seq_len)
    print "全断片数は#{dna_seq_frag_ary.length}です\n"

    i = 0
    sp = 0
    
    #断片ごとに解析    
    name = flat_file_name.chomp + end_name
    w_file = File.open("#{name}", "w")

    dna_seq_frag_ary.each do |dna_seq_ary|
      
      #Nが断片の10%以下かどうか
      if contentNunder10(dna_seq_ary)
        
        #連続塩基頻度解析
        oligonucleotide_freq = calculate_oligonucleotide_freq(dna_seq_ary, word_len, same_pattern)
       
        #コメント行の作成
        if (sp + WINDOW_SIZE) > dna_seq_len          
          comment = i.to_s + " [" + (dna_seq_len - WINDOW_SIZE).to_s + "-" + (dna_seq_len - 1).to_s + "] < " + flat_name
        else
          comment = i.to_s + " [" + sp.to_s + "-" + (sp + WINDOW_SIZE - 1).to_s + "] < " + flat_name
        end

        i += 1
        sp += STEP_SIZE
        
        #出力
        output(flat_name, comment, oligonucleotide_freq, word_pattern, w_file)
#        output(first_line, comment, oligonucleotide_freq, word_pattern, w_file)
#        print "#{i}個の断片の処理を完了\n"
      else
        sp += STEP_SIZE
      end      
    end
    w_file.close
  else
    print("\nError : seq less windowsize
入力ファイルの塩基配列長が断片化サイズ以下になっています
入力ファイルを確認するか，
freq_analysis.rbの設定部分のWINDOW_SIZEを変更してください\n\n") #配列がWINDOWSIZE以下だった場合 
  end
end



##########(塩基配列を断片化するメソッド)##########
def fragmentation(dna_seq_str, dna_len = nil)
  print "塩基配列の断片化を開始します\n"

  dna_seq_frag_ary = []        # ウインドウサイズで断片化された配列の格納
  
  #1文字ずつ入れる
  dna_seq_ary = dna_seq_str.split("")
  
  if !dna_len
    dna_len = dna_seq_str.length
  end
  
  sp = 0
  ep = sp + WINDOW_SIZE - 1  
  i = 0
  
  while sp < dna_len
    
    if ep > dna_len - 1
      # 末尾までウインドウサイズ分ないときはちょうどwindowsize分になるように調節
      sp = dna_len - WINDOW_SIZE
      ep = dna_len -1
      dna_seq_frag_ary[i] = dna_seq_ary[sp..ep]
    else
      dna_seq_frag_ary[i] = dna_seq_ary[sp..ep]
    end
    
    sp += STEP_SIZE
    ep += STEP_SIZE
    i += 1

  end
  
  return dna_seq_frag_ary
  
end



##########(Nのパーセンテージによって単語を除外するメソッド1)##########
def contentNover10(dna_seq_ary)
  
  count = 0
  dna_len = dna_seq_ary.length
  
  dna_seq_ary.each do |nc|
    if nc !~ /A|a|T|t|G|g|C|c/
      count += 1
    end
  end
  
  #===========================Nのパーセンテージ==========================
  if (count.to_f / dna_len.to_f) >= NPAR
  #======================================================================
    return true
  else
    return false
  end

end



##########(Nのパーセンテージによって単語を除外するメソッド2)##########
def contentNunder10(dna_seq_ary)
  !(contentNover10(dna_seq_ary))
end



##########(頻度計算を行うメソッド)###########
def calculate_oligonucleotide_freq(dna_seq_ary, word_len, pattern)
  onc = oligonucleotide_count(dna_seq_ary, word_len)  # 縮退四連の頻度計算  
  dna_len = dna_seq_ary.length
  
  #断片中の単語数(全四連塩基の数)
  word_total = dna_len - (word_len - 1)

  onc_freq_ConsiderlateSingleNC = Hash.new(0)
  
  #Nの混じった単語は使わないので削除して断片中の単語数を正しい値に
  onc.each do |word, count|
    if word !~ /[A|a|T|t|G|g|C|c]{#{word_len}}/
      word_total -= count
      onc.delete(word)
    end
  end
  
  
  #同じ単語パターンの時はカウント+1
  onc.each do |word, count|
    i = 0
    while line = pattern[i]
      i += 1
      if word == line
        word_total += count
      end
    end
  end
  
  
  onc.each{|word, count,|

    #計算実行
    case CALC      
    when "100hindo"   #100分率頻度計算

      onc_freq = count.to_f / word_total
      onc_freq_ConsiderlateSingleNC[word] = onc_freq * 100
            
    when "hindo"   #1分率頻度計算

      onc_freq = count.to_f / word_total
      onc_freq_ConsiderlateSingleNC[word] = onc_freq

    when "dosu"   #度数計算

      onc_freq = count.to_f
      onc_freq_ConsiderlateSingleNC[word] = onc_freq

    end


  }
  
  return onc_freq_ConsiderlateSingleNC
end



##########(単語をカウントするメソッド)##########
def oligonucleotide_count(dna_seq_ary, word_len)

  word = ""
  onc_count = Hash.new(0)
  i = 0

  while dna_seq_ary[i + word_len - 1]
    word_len.times do |j|
      word << dna_seq_ary[i+j]     # wordに四連塩基が格納される
    end
    
    onc_count[word] += 1
    word = ""
    i += 1
  end
  
  return onc_count
end



##########(出力をするメソッド)##########
def output(first_line, second_line, freq_data, word_pattern, w_file)
  w_file.print ">#{first_line}\n"
#  w_file.print "#{first_line}\n"
  w_file.print "#{second_line}\n"
  
  word_pattern.each{|ptn|
    if ptn =~ /\+/
      tmp = ptn.split('+')
      tmp[1].chomp
      w_file.print(freq_data[tmp[0]] + freq_data[tmp[1]], ' ')
    else
      w_file.print(freq_data[ptn], ' ')
    end
  }
  w_file.print("\n")
end



###########(main)############
pattern_file_name = ARGV[1]

input_file = ARGV[0]  
oligonucleotide_freq(input_file, pattern_file_name, end_name)
print "終了しました\n"
